/*****************************************************************************

NAME:
   bogofilter.c -- detect spam and bogons presented on standard input.

AUTHOR:
   Eric S. Raymond <esr@thyrsus.com>

THEORY:
   This is Paul Graham's variant of Bayes filtering described at 

	http://www.paulgraham.com/spam.html

I do the lexical analysis slightly differently, however.

******************************************************************************/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include "bogofilter.h"

// like the stock Judy error macro, except with exit(2)
#define JUDYERROR(CallerFile, CallerLine, JudyFunc, JudyErrno, JudyErrID) \
	{							\
	    (void) fprintf(stderr, "File '%s', line %d: %s(), "	\
			   "bogofilter: JU_ERRNO_* == %d, ID == %d\n",	\
			   CallerFile, CallerLine,		\
			   JudyFunc, JudyErrno, JudyErrID);	\
	    exit(2);						\
	}


#include <Judy.h>

// implementation details
#define HEADER		"# bogofilter wordlist (format version A): %lu\n"

#define DEVIATION(n)	fabs((n) - 0.5f)	// deviation from average

#define max(x, y)	(((x) > (y)) ? (x) : (y))
#define min(x, y)	(((x) < (y)) ? (x) : (y))

wordlist_t good_list	= {"good", NULL, 0, 0, NULL};
wordlist_t bad_list	= {"bad", NULL, 0, 0, NULL};

#define	PLURAL(count) ((count == 1) ? "" : "s")

static void dump(wordlist_t *words, FILE *fpout)
{
    char	tokenbuffer[BUFSIZ];
    void	**loc;

    tokenbuffer[0]='\0';
    for (loc  = JudySLFirst(words->list, tokenbuffer, 0);
	 loc != (void *) NULL;
	 loc  = JudySLNext(words->list, tokenbuffer, 0))
    {
	// reminder -- this goes to disk, not up a socket
	(void) fprintf(fpout, "%s %lu\n", tokenbuffer, (*((PWord_t) loc)));
    }

}

static void strlwr(char* s)
{
    char c;
    while((c = *s) != 0)
	*s++ = tolower(c);
}

static int increment(char *word,  int incr, wordlist_t *list)
/* increment a word usage count in the specified list */
{
    void  ** PPValue;			// associated with Index.
    void  *  PArray = (Pvoid_t) NULL;	// JudySL array.
    JError_t JError;                    // Judy error structure

    if ((PPValue = JudySLIns(&(list->list), word, &JError)) == PPJERR)
	return(-1);

    list->modcount++;
    if (incr > 0) {
	(*((PWord_t) PPValue)) += incr;	// increment usage count.
    } else {
	Word_t old = *((PWord_t) PPValue);
	incr = -incr;
	if ((Word_t)old < incr) {
	    incr = old;
	}
	*((PWord_t) PPValue) = old - (Word_t)incr;
    }
}

int getcount(char *word, wordlist_t *list)
/* get the count associated with a given word in a list */
{
    Word_t * PValue;                     // JudySL array element

    JSLG(PValue, list->list, word);
    if (PValue == NULL)
	return(0);
    else
	return *PValue;
}

void register_words(int fdin, wordlist_t *list, wordlist_t *other, int fdout)
// tokenize text on stdin and register it to  a specified list
// and possibly out of another list
{
    int	tok, wordcount, msgcount;
    void  **PPValue;			// associated with Index.
    void  *PArray = (Pvoid_t) NULL;	// JudySL array.
    JError_t JError;                    // Judy error structure
    void	**loc;
    char	tokenbuffer[BUFSIZ];

    // Grab tokens from the lexical analyzer into our own private Judy array
    yyin = fdopen(fdin, "r");
    wordcount = msgcount = 0;
    for (;;)
    {
	tok = get_token();

	if (tok != FROM && tok != 0)
	{
	    // Ordinary word, stash in private per-message array.
	    strlwr(yytext);
	    if ((PPValue = JudySLIns(&PArray, yytext, &JError)) == PPJERR)
		return;
	    (*((PWord_t) PPValue))++;
	    wordcount++;
	}
	else
	{
	    // End of message. Update message counts.
	    if (tok == FROM || (tok == 0 && msgcount == 0))
	    {
		list->msgcount++;
		msgcount++;
		if (other && other->msgcount > 0)
		    other->msgcount--;
	    }

	    // We copy the incoming words into their own per-message array
	    // in order to be able to cap frequencies.
	    tokenbuffer[0]='\0';
	    for (loc  = JudySLFirst(PArray, tokenbuffer, 0);
		 loc != (void *) NULL;
		 loc  = JudySLNext(PArray, tokenbuffer, 0))
	    {
		int freq	= (*((PWord_t) loc));

		if (freq > MAX_REPEATS)
		    freq = MAX_REPEATS;

		increment(tokenbuffer, freq, list);
		if (other)
		    increment(tokenbuffer, -freq, other);
	    }
	    JudySLFreeArray(&PArray, &JError);
	    PArray = (Pvoid_t)NULL;

	    if (fdout != -1)
		sockprintf(fdout, "# %d words\n", wordcount);

	    // Want to process EOF, *then* drop out
	    if (tok == 0)
		break;
	}
    }
}

// Nothing below this point knows about the Judy interface

int read_list(char *file, wordlist_t *list)
/* initialize one of the associative arrays from a file */
/* return 0 if file read was successful, and 1 if it was unsuccessful.
   note that failure due to missing file is ok if we're planning
   on adding to the list. */
{
    char	tokenbuffer[BUFSIZ];
    FILE	*infp;
    Word_t	count;

    list->file = strdup(file);
    infp = fopen(file, "r");
    if (infp == NULL)
	return(-1);

    fscanf(infp, HEADER, &list->msgcount);
    while (fscanf(infp, "%s %lu", tokenbuffer, &count) == 2)
	increment(tokenbuffer, count, list);

    fclose(infp);
    return(0);
}

static int write_list(wordlist_t *list)
/* dump the state of one of the arrays to a specified file */
{
    FILE	*outfp;
    struct flock	lockall = {F_RDLCK, SEEK_SET, 0L, 0L};

    outfp = fopen(list->file, "a+");
    if (outfp == NULL)
	return(-1);

    /* lock for read and then truncate in order to avoid a race
     * condition with bogofilter instances that start up to check mail 
     * as the file is being written.
     */
    lockall.l_type = F_RDLCK;
    fcntl(fileno(outfp), F_SETLKW, &lockall);
    ftruncate(fileno(outfp), 0L);
    (void) fprintf(outfp, HEADER, list->msgcount);
    dump(list, outfp);
    fclose(outfp);
    lockall.l_type = F_UNLCK;
    fcntl(fileno(outfp), F_SETLKW, &lockall);

    return(0);
}

void write_lists(void)
// flush all lists with modified entries
{
    if (good_list.modcount)
	write_list(&good_list);
    if (bad_list.modcount)
	write_list(&bad_list);
}


#ifdef __UNUSED__
void logprintf(const char *fmt, ... )
// log data from server
{
    char buf[BUFSIZ];
    va_list ap;
    int fd;

    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    
    fd=open("/tmp/bogolog", O_RDWR|O_CREAT|O_APPEND,0700);
    write(fd,buf,strlen(buf));
    close(fd);
}
#endif // __UNUSED__

void statdump(const bogostat_t *pb, int fd)
// Dump the contents of a statistics structure
{
    char buf[BUFSIZ];
    const discrim_t *pp;

    sockprintf(fd, "# Spamicity: %f\n", pb->spamicity);
    for (pp = pb->extrema; pp < pb->extrema + KEEPERS; pp++)
	if (pp->key[0])
	    sockprintf(fd, "# '%s' -> %f\n", pp->key, pp->prob);
}

bogostat_t *bogofilter(int fd)
/* evaluate text for spamicity */
{
    double prob, product, invproduct;
    double goodness, badness, slotdev, hitdev;    
    int tok;
#ifdef NON_EQUIPROBABLE
    // There is an argument that we should by by number of *words* here.
    double	msg_prob = (bad_list.msgcount / good_list.msgcount);
#endif // NON_EQUIPROBABLE

    static bogostat_t stats;
    discrim_t *pp, *hit;

    for (pp = stats.extrema; pp < stats.extrema+sizeof(stats.extrema)/sizeof(*stats.extrema); pp++)
    {
	pp->prob = 0.5f;
	pp->key[0] = '\0';
    }

    yyin = fdopen(fd, "r");
    while ((tok = get_token()) != 0)
    {
	double dev;

	strlwr(yytext);
	goodness = getcount(yytext, &good_list);
	badness  = getcount(yytext, &bad_list);

	// Paul Graham's original formula:
	// 
	// (let ((g (* 2 (or (gethash word good) 0))) 
	//      (b (or (gethash word bad) 0)))
	//  (unless (&lt; (+ g b) 5) 
	//   (max .01 (min .99 
	//  	    (double (/ 
	// 		    (min 1 (/ b nbad)) 
	// 		    (+ (min 1 (/ g ngood)) (min 1 (/ b nbad)))))))))
	// This assumes that spam and non-spam are equiprobable.
	goodness *= GOOD_BIAS;
	if (goodness + badness < MINIMUM_FREQ)
#ifdef NON_EQUIPROBABLE
	    // In the absence of evidence, the probability that a new word
	    // will be spam is the historical ratio of spam words to
	    // nonspam words.
	    prob = msg_prob;
#else
	    prob = UNKNOWN_WORD;
#endif // NON_EQUIPROBABLE
	else
	{
	    register double pb = min(1, (badness / bad_list.msgcount));
	    register double pg = min(1, (goodness / good_list.msgcount));

#ifdef NON_EQUIPROBABLE
	    prob = (pb * msg_prob) / ((pg * (1 - msg_prob)) + (pb * msg_prob));
#else
	    prob = pb / (pg + pb);
#endif // NON_EQUIPROBABLE
	    prob = min(prob, 0.99);
	    prob = max(prob, 0.01);
	}

	// update the list of tokens with maximum deviation
	dev = DEVIATION(prob);
        hit = NULL;
        hitdev=0;
	for (pp = stats.extrema; pp < stats.extrema+sizeof(stats.extrema)/sizeof(*stats.extrema); pp++)
        {
	    // don't allow duplicate tokens in the stats.extrema
	    if (pp->key && strcmp(pp->key, yytext)==0)
            {
                hit=NULL;
		break;
            }
	    else 
	    {
                slotdev=DEVIATION(pp->prob);
                if (dev>slotdev && dev>hitdev)
                {
                    hit=pp;
		    hitdev=slotdev;
                }
            }
        }
        if (hit) 
	{ 
	    hit->prob = prob;
	    strncpy(hit->key, yytext, MAXWORDLEN);
	}
    }

    // Bayes' theorem.
    // For discussion, see <http://www.mathpages.com/home/kmath267.htm>.
    product = invproduct = 1.0f;
    for (pp = stats.extrema; pp < stats.extrema+sizeof(stats.extrema)/sizeof(*stats.extrema); pp++)
	if (pp->prob == 0)
	    break;
    	else
	{
	    product *= pp->prob;
	    invproduct *= (1 - pp->prob);
	}
    stats.spamicity = product / (product + invproduct);

    return(&stats);
}

/* Done */
